# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from enum import auto, IntEnum, Enum
from typing import TYPE_CHECKING, Tuple, Any, Optional

import iso3166

from .sorted_list import Sortable
from ...base.exception import AccessDeniedException
from ...base.type_converter_templates import ConstantKeys
from ...icon_constant import PRepGrade, PRepStatus, PenaltyReason, Revision, PRepFlag
from ...utils.msgpack_for_db import MsgPackForDB

if TYPE_CHECKING:
    from iconservice.base.address import Address


class PRepDictType(Enum):
    FULL = auto()  # getPRep
    ABRIDGED = auto()  # getPReps


class PRep(Sortable):
    PREFIX: bytes = b"prep"
    _VERSION: int = 2
    _UNKNOWN_COUNTRY = iso3166.Country(u"Unknown", "ZZ", "ZZZ", "000", u"Unknown")

    class Index(IntEnum):
        VERSION = 0

        # Version: 0
        ADDRESS = auto()
        STATUS = auto()
        GRADE = auto()
        NAME = auto()
        COUNTRY = auto()
        CITY = auto()
        EMAIL = auto()
        WEBSITE = auto()
        DETAILS = auto()
        P2P_ENDPOINT = auto()
        IREP = auto()
        IREP_BLOCK_HEIGHT = auto()
        LAST_GENERATE_BLOCK_HEIGHT = auto()
        BLOCK_HEIGHT = auto()
        TX_INDEX = auto()
        TOTAL_BLOCKS = auto()
        VALIDATED_BLOCKS = auto()

        # Version: 1
        PENALTY = auto()
        UNVALIDATED_SEQUENCE_BLOCKS = auto()

        NODE_ADDRESS = auto()

        # Unused
        SIZE = auto()

    def __init__(
            self, address: 'Address',
            *,
            flags: 'PRepFlag' = PRepFlag.NONE,
            status: 'PRepStatus' = PRepStatus.ACTIVE,
            penalty: 'PenaltyReason' = PenaltyReason.NONE,
            grade: 'PRepGrade' = PRepGrade.CANDIDATE,
            name: str = "",
            country: str = "",
            city: str = "Unknown",
            email: str = "",
            website: str = "",
            details: str = "",
            p2p_endpoint: str = "",
            irep: int = 0,
            irep_block_height: int = 0,
            last_generate_block_height: int = -1,
            stake: int = 0,
            delegated: int = 0,
            block_height: int = 0,
            tx_index: int = 0,
            total_blocks: int = 0,
            validated_blocks: int = 0,
            unvalidated_sequence_blocks: int = 0,
            node_address: Optional['Address'] = None,

    ):
        """
        Main PRep: top 1 ~ 22 preps in descending order by delegated amount
        Sub PRep: 23 ~ 100 preps
        PRep: All PReps including Main PRep and Sub PRep

        :param address:
        :param flags:
        :param status:
        :param penalty:
        :param grade:
        :param name:
        :param country: alpha3 country code (ISO3166)
        :param city:
        :param email:
        :param website:
        :param details:
        :param p2p_endpoint:
        :param irep:
        :param irep_block_height:
        :param stake:
        :param delegated:
        :param block_height:
        :param tx_index:
        :param total_blocks:
        :param validated_blocks:
        :param unvalidated_sequence_blocks
        :param node_address
        """
        # key
        self._address: 'Address' = address

        # flags
        self._flags: 'PRepFlag' = flags

        self._stake: int = stake
        self._delegated: int = delegated

        # status
        self._status: 'PRepStatus' = status
        self._penalty: 'PenaltyReason' = penalty
        self._grade: 'PRepGrade' = grade

        # registration info
        self._name: str = name
        self._country: 'iso3166.Country' = self._get_country(country)
        self._city: str = city
        self._email: str = email
        self._website: str = website
        self._details: str = details
        # information required for PRep Consensus
        self._p2p_endpoint: str = p2p_endpoint
        # Governance Variables
        self._irep: int = irep
        self._irep_block_height: int = irep_block_height

        # The height of the last block which a P-Rep generated
        self._last_generate_block_height: int = last_generate_block_height

        # registration time
        self._block_height: int = block_height
        self._tx_index: int = tx_index

        # stats
        self._total_blocks: int = total_blocks
        self._validated_blocks: int = validated_blocks
        self._unvalidated_sequence_blocks: int = unvalidated_sequence_blocks

        self._is_frozen: bool = False

        # node key
        self._node_address = node_address if node_address else address

    def is_flags_on(self, flags: 'PRepFlag') -> bool:
        return (self._flags & flags) == flags

    def is_dirty(self) -> bool:
        """It returns True if any PRepFlag is True

        :return:
        """
        return bool(self._flags & PRepFlag.ALL)

    @property
    def flags(self) -> 'PRepFlag':
        return self._flags

    @property
    def status(self) -> 'PRepStatus':
        return self._status

    @status.setter
    def status(self, value: 'PRepStatus'):
        self._set_property(name="_status", new_value=value, flags=PRepFlag.STATUS)

    @property
    def penalty(self) -> 'PenaltyReason':
        return self._penalty

    @penalty.setter
    def penalty(self, value: 'PenaltyReason'):
        self._set_property(name="_penalty", new_value=value, flags=PRepFlag.PENALTY)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._set_property(name="_name", new_value=value, flags=PRepFlag.NAME)

    @property
    def country(self) -> str:
        return self._country.alpha3

    @country.setter
    def country(self, alpha3_country_code: str):
        value = self._get_country(alpha3_country_code)
        self._set_property(name="_country", new_value=value, flags=PRepFlag.COUNTRY)

    @classmethod
    def _get_country(cls, alpha3_country_code: str) -> 'iso3166.Country':
        return iso3166.countries_by_alpha3.get(
            alpha3_country_code.upper(), cls._UNKNOWN_COUNTRY)

    @property
    def city(self) -> str:
        return self._city

    @city.setter
    def city(self, value: str):
        self._set_property(name="_city", new_value=value, flags=PRepFlag.CITY)

    @property
    def email(self) -> str:
        return self._email

    @email.setter
    def email(self, value: str):
        self._set_property(name="_email", new_value=value, flags=PRepFlag.EMAIL)

    @property
    def website(self) -> str:
        return self._website

    @website.setter
    def website(self, value: str):
        self._set_property(name="_website", new_value=value, flags=PRepFlag.WEBSITE)

    @property
    def details(self) -> str:
        return self._details

    @details.setter
    def details(self, value: str):
        self._set_property(name="_details", new_value=value, flags=PRepFlag.DETAILS)

    @property
    def p2p_endpoint(self) -> str:
        return self._p2p_endpoint

    @p2p_endpoint.setter
    def p2p_endpoint(self, value: str):
        self._set_property(name="_p2p_endpoint", new_value=value, flags=PRepFlag.P2P_ENDPOINT)

    def is_suspended(self) -> bool:
        """The suspended P-Rep cannot serve as Main P-Rep during this term

        :return:
        """
        return self._penalty == PenaltyReason.BLOCK_VALIDATION

    def is_electable(self) -> bool:
        """Returns whether this P-Rep can be elected as a Main P-Rep or Sub P-Rep

        :return:
        """
        return self._status == PRepStatus.ACTIVE and self._penalty == PenaltyReason.NONE

    @property
    def grade(self) -> 'PRepGrade':
        """The grade of P-Rep
        0: MAIN
        1: SUB
        2: CANDIDATE

        :return:
        """
        return self._grade

    @grade.setter
    def grade(self, value: 'PRepGrade'):
        self._set_property(name="_grade", new_value=value, flags=PRepFlag.GRADE)

    def update_block_statistics(self, is_validator: bool):
        """Update the block validation statistics of P-Rep

        :param is_validator: If this P-Rep validates a block, then it is True
        """
        self._check_access_permission()

        self._set_property("_total_blocks", self._total_blocks + 1, PRepFlag.TOTAL_BLOCKS, False)

        if is_validator:
            self._set_property("_validated_blocks", self._validated_blocks + 1, PRepFlag.VALIDATED_BLOCKS, False)
            self._set_property("_unvalidated_sequence_blocks", 0, PRepFlag.UNVALIDATED_SEQUENCE_BLOCKS, False)
        else:
            self._set_property(
                "_unvalidated_sequence_blocks",
                self._unvalidated_sequence_blocks + 1,
                PRepFlag.UNVALIDATED_SEQUENCE_BLOCKS,
                False)

    def reset_block_validation_penalty(self):
        """Reset block validation penalty and
        unvalidated sequence blocks before the next term begins

        :return:
        """
        self._check_access_permission()
        self._set_property("_penalty", PenaltyReason.NONE, PRepFlag.PENALTY, False)
        self._set_property("_unvalidated_sequence_blocks", 0, PRepFlag.UNVALIDATED_SEQUENCE_BLOCKS, False)

    @property
    def total_blocks(self) -> int:
        return self._total_blocks

    @property
    def validated_blocks(self) -> int:
        return self._validated_blocks

    @property
    def block_validation_proportion(self) -> int:
        """Percent without fraction

        :return:
        """
        if self._total_blocks == 0:
            return 0

        return self._validated_blocks * 100 // self._total_blocks

    @property
    def unvalidated_sequence_blocks(self) -> int:
        return self._unvalidated_sequence_blocks

    @property
    def address(self) -> 'Address':
        return self._address

    @property
    def node_address(self) -> 'Address':
        return self._node_address

    @node_address.setter
    def node_address(self, value: 'Address'):
        self._set_property(name="_node_address", new_value=value, flags=PRepFlag.NODE_ADDRESS)

    @property
    def stake(self) -> int:
        return self._stake

    @stake.setter
    def stake(self, value: int):
        """stake is set to account.stake

        :param value:
        :return:
        """
        assert value >= 0
        self._set_property(name="_stake", new_value=value, flags=PRepFlag.STAKE)

    @property
    def delegated(self) -> int:
        return self._delegated

    @delegated.setter
    def delegated(self, value: int):
        """delegated is set to account.delegated_amount

        :param value:
        :return:
        """
        assert value >= 0
        self._set_property(name="_delegated", new_value=value, flags=PRepFlag.DELEGATED)

    @property
    def irep(self) -> int:
        return self._irep

    @property
    def irep_block_height(self) -> int:
        return self._irep_block_height

    @property
    def last_generate_block_height(self) -> int:
        return self._last_generate_block_height

    @last_generate_block_height.setter
    def last_generate_block_height(self, value: int):
        assert value >= 0
        self._set_property(name="_last_generate_block_height",
                           new_value=value,
                           flags=PRepFlag.LAST_GENERATE_BLOCK_HEIGHT)

    @property
    def block_height(self) -> int:
        return self._block_height

    @property
    def tx_index(self) -> int:
        return self._tx_index

    @classmethod
    def make_key(cls, address: 'Address') -> bytes:
        return cls.PREFIX + address.to_bytes_including_prefix()

    def is_frozen(self) -> bool:
        return self._is_frozen

    def freeze(self):
        """Make all member variables immutable
        """
        self._is_frozen = True
        self._flags = PRepFlag.NONE

    def set(self,
            *,
            name: str = None,
            country: str = None,
            city: str = None,
            email: str = None,
            website: str = None,
            details: str = None,
            p2p_endpoint: str = None,
            node_address: 'Address' = None):
        """Update PRep properties on processing setPRep JSON-RPC API
        Not allowed to update some properties which can affect PRep order or are immutable

        :param name:
        :param country: alpha3 country code
        :param city:
        :param email:
        :param website:
        :param details:
        :param p2p_endpoint:
        :param node_address:
        """
        self._check_access_permission()

        kwargs = {
            "name": name,
            "country": country,
            "city": city,
            "email": email,
            "website": website,
            "details": details,
            "p2p_endpoint": p2p_endpoint,
            "node_address": node_address
        }

        for key, value in kwargs.items():
            if value is not None:
                setattr(self, key, value)

    def set_irep(self, irep: int, block_height: int):
        """Set incentive rep

        :param irep:
        :param block_height: block height when irep is set
        :return:
        """
        self._check_access_permission()

        self._set_property("_irep", irep, PRepFlag.IREP, False)
        self._set_property("_irep_block_height", block_height, PRepFlag.IREP_BLOCK_HEIGHT, False)

    def __gt__(self, other: 'PRep') -> bool:
        return self.order() > other.order()

    def __lt__(self, other: 'PRep') -> bool:
        return self.order() < other.order()

    def order(self) -> Tuple[int, int, int]:
        """Returns tuple used for sorting preps in descending order

        :return: delegated, block_height, tx_index
        """
        return -self._delegated, self._block_height, self._tx_index

    def to_bytes(self, revision: int) -> bytes:
        if revision >= Revision.DIVIDE_NODE_ADDRESS.value:
            version: int = 2
        elif revision >= Revision.DECENTRALIZATION.value:
            version: int = 1
        else:
            version: int = 0

        data = [
            version,
            self.address,
            self.status.value,
            self.grade.value,
            self.name,
            self.country,
            self.city,
            self.email,
            self.website,
            self.details,
            self.p2p_endpoint,

            self._irep,
            self._irep_block_height,

            self._last_generate_block_height,

            self._block_height,
            self._tx_index,

            self._total_blocks,
            self._validated_blocks,
        ]

        if version >= 1:
            data.extend((self.penalty.value, self._unvalidated_sequence_blocks))

        if version >= 2:
            data.append(self._node_address)

        return MsgPackForDB.dumps(data)

    @classmethod
    def from_bytes(cls, data: bytes) -> 'PRep':
        items: list = MsgPackForDB.loads(data)
        version: int = items[cls.Index.VERSION]

        if version == 0:
            items.extend((PenaltyReason.NONE, 0))

        if version < 2:
            items.append(items[cls.Index.ADDRESS])

        return PRep(
            # version 0
            address=items[cls.Index.ADDRESS],
            status=PRepStatus(items[cls.Index.STATUS]),
            grade=PRepGrade(items[cls.Index.GRADE]),
            name=items[cls.Index.NAME],
            country=items[cls.Index.COUNTRY],
            city=items[cls.Index.CITY],
            email=items[cls.Index.EMAIL],
            website=items[cls.Index.WEBSITE],
            details=items[cls.Index.DETAILS],
            p2p_endpoint=items[cls.Index.P2P_ENDPOINT],
            irep=items[cls.Index.IREP],
            irep_block_height=items[cls.Index.IREP_BLOCK_HEIGHT],
            last_generate_block_height=items[cls.Index.LAST_GENERATE_BLOCK_HEIGHT],
            block_height=items[cls.Index.BLOCK_HEIGHT],
            tx_index=items[cls.Index.TX_INDEX],
            total_blocks=items[cls.Index.TOTAL_BLOCKS],
            validated_blocks=items[cls.Index.VALIDATED_BLOCKS],

            # version 1
            penalty=PenaltyReason(items[cls.Index.PENALTY]),
            unvalidated_sequence_blocks=items[cls.Index.UNVALIDATED_SEQUENCE_BLOCKS],

            # version 2
            node_address=items[cls.Index.NODE_ADDRESS],
        )

    @staticmethod
    def from_dict(address: 'Address', data: dict, block_height: int, tx_index: int) -> 'PRep':
        """Create a PRep instance from data included in registerPRep JSON-RPC request

        :param address:
        :param data:
        :param block_height:
        :param tx_index:
        :return:
        """
        return PRep(
            flags=PRepFlag.NONE,
            address=address,
            status=PRepStatus.ACTIVE,
            grade=PRepGrade.CANDIDATE,

            # Optional items
            name=data.get(ConstantKeys.NAME, ""),
            country=data.get(ConstantKeys.COUNTRY, ""),
            city=data.get(ConstantKeys.CITY, ""),
            email=data.get(ConstantKeys.EMAIL, ""),
            website=data.get(ConstantKeys.WEBSITE, ""),
            details=data.get(ConstantKeys.DETAILS, ""),

            # Required items
            p2p_endpoint=data[ConstantKeys.P2P_ENDPOINT],
            irep=0,
            irep_block_height=block_height,

            # Registration time
            block_height=block_height,
            tx_index=tx_index,

            # node key
            node_address=data.get(ConstantKeys.NODE_ADDRESS)
        )

    def to_dict(self, dict_type: 'PRepDictType') -> dict:
        """Returns the P-Rep information in dict format

        :param dict_type: FULL(getPRep), ABRIDGED(getPReps)
        :return:
        """
        data = {
            "address": self._address,
            "status": self._status.value,
            "penalty": self._penalty.value,
            "grade": self.grade.value,
            "name": self.name,
            "country": self.country,
            "city": self.city,
            "stake": self._stake,
            "delegated": self._delegated,
            "totalBlocks": self._total_blocks,
            "validatedBlocks": self._validated_blocks,
            "unvalidatedSequenceBlocks": self._unvalidated_sequence_blocks,
            "irep": self._irep,
            "irepUpdateBlockHeight": self._irep_block_height,
            "lastGenerateBlockHeight": self._last_generate_block_height,
            "blockHeight": self._block_height,
            "txIndex": self._tx_index,
            "nodeAddress": self._node_address,
        }
        # TODO Revision: remove irep, irepUpdateBlockHeight after SET_IREP_VIA_NETWORK_PROPOSAL

        if dict_type == PRepDictType.FULL:
            data[ConstantKeys.EMAIL] = self.email
            data[ConstantKeys.WEBSITE] = self.website
            data[ConstantKeys.DETAILS] = self.details
            data[ConstantKeys.P2P_ENDPOINT] = self.p2p_endpoint

        return data

    def __str__(self) -> str:
        info: dict = self.to_dict(PRepDictType.FULL)
        return str(info)

    def copy(self) -> 'PRep':
        prep = copy.copy(self)
        prep._is_frozen = False
        prep._flags = PRepFlag.NONE

        return prep

    def _check_access_permission(self):
        if self.is_frozen():
            raise AccessDeniedException("P-Rep access denied")

    def _set_property(self, name: str, new_value: Any, flags: 'PRepFlag', check_permission: bool = True):
        if check_permission:
            self._check_access_permission()

        old_value = getattr(self, name)
        if old_value != new_value:
            setattr(self, name, new_value)
            self._flags |= flags



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import shutil
import zipfile

from ..base.exception import InvalidPackageException
from ..icon_constant import Revision, PACKAGE_JSON_FILE


class IconScoreDeployer(object):

    @staticmethod
    def deploy(path: str, data: bytes, revision: int = 0):
        """Deploy SCORE; Stores SCORE on the root path

        :param path: the path of directory where score is deployed
        :param data: Bytes of the zip file.
        :param revision: Revision num
        """
        shutil.rmtree(path, ignore_errors=True)
        os.makedirs(path)

        file_info_generator = IconScoreDeployer._extract_files_gen(data, revision)
        for name, file_info, parent_dir in file_info_generator:
            if not os.path.exists(os.path.join(path, parent_dir)):
                os.makedirs(os.path.join(path, parent_dir))
            with file_info as file_info_context, open(os.path.join(path, name), 'wb') as dest:
                contents = file_info_context.read()
                dest.write(contents)

    @staticmethod
    def _extract_files_gen(data: bytes, revision: int = 0):
        """
        Reads all files from the depth lower than where the file 'package.json' is and make the generator.
        The generator has tuples with a filename, file info, parent dir.
        When revision is 2 or more, this method is used.

        :param data: Bytes of the zip file.
        :param revision: Revision num.
        """

        try:
            with zipfile.ZipFile(io.BytesIO(data)) as memory_zip:
                memory_zip_infolist = memory_zip.infolist()
                common_prefix = ""
                has_package = False
                # Finds the depth having the file 'package.json'.
                for zip_info in memory_zip_infolist:
                    file_path = zip_info.filename
                    file_name = os.path.basename(file_path)
                    if PACKAGE_JSON_FILE == file_name:
                        common_prefix = os.path.dirname(file_path)
                        has_package = True
                        break

                if revision >= Revision.THREE.value and has_package is False:
                    raise InvalidPackageException("package.json not found")

                for zip_info in memory_zip_infolist:
                    with memory_zip.open(zip_info) as file:
                        file_path = zip_info.filename
                        if (
                                file_path.startswith(common_prefix)
                                and not zip_info.is_dir()
                                and file_path.find('__MACOSX') < 0
                                and file_path.find('__pycache__') < 0
                                and not file_path.startswith('.')
                                and file_path.find('/.') < 0
                        ):
                            if revision >= Revision.THREE.value:
                                file_path = os.path.relpath(file_path, common_prefix)
                                parent_directory = os.path.dirname(file_path)
                                if file_path:
                                    yield file_path, file, parent_directory
                            else:
                                # legacy for revision 2
                                legacy_common_prefix = f"{common_prefix}/"
                                file_path = file_path.replace(legacy_common_prefix, '')
                                parent_directory = os.path.dirname(file_path)
                                if file_path and file_path[-1] != '/':
                                    yield file_path, file, parent_directory
        except zipfile.BadZipFile:
            raise InvalidPackageException("Bad zip file.")
        except zipfile.LargeZipFile:
            raise InvalidPackageException("Zip file is too Large.")
        except Exception as e:
            raise InvalidPackageException(f'Error raised from extract_files_gen: {e}')

    @staticmethod
    def deploy_legacy(path: str, data: bytes):
        """Install score.
        Use 'address', 'block_height', and 'transaction_index' to specify the path where 'Score' will be installed.

        :param path: the path of directory where score is deployed
        :param data: The byte value of the zip file.
        """
        shutil.rmtree(path, ignore_errors=True)
        if not os.path.exists(path):
            os.makedirs(path)

        file_info_generator = IconScoreDeployer._extract_files_gen_legacy(data)
        for name, file_info, parent_directory in file_info_generator:
            if not os.path.exists(os.path.join(path, parent_directory)):
                os.makedirs(os.path.join(path, parent_directory))
            with file_info as file_info_context, open(os.path.join(path, name), 'wb') as dest:
                contents = file_info_context.read()
                dest.write(contents)

    @staticmethod
    def _extract_files_gen_legacy(data: bytes):
        """Yield (filename, file_information, parent_directory_name) tuple.

        :param data: The byte value of the zip file.
        :return:
        """
        try:
            with zipfile.ZipFile(io.BytesIO(data)) as memory_zip:
                memory_zip_infolist = memory_zip.infolist()
                memory_zip_files_path_gen = (path.filename for path in memory_zip_infolist)
                common_path_len = len(os.path.commonpath(memory_zip_files_path_gen))
                start_index = common_path_len
                if common_path_len == 0:
                    start_index = -1
                for zip_info in memory_zip.infolist():
                    with memory_zip.open(zip_info) as file:
                        file_path = zip_info.filename[start_index + 1:]
                        file_name_start_index = file_path.rfind('/')
                        parent_directory = file_path[:file_name_start_index]
                        if file_path.find('__MACOSX') != -1:
                            continue
                        if file_path.find('__pycache__') != -1:
                            continue
                        if file_name_start_index == len(file_path) - 1:
                            # continue when 'file_path' is a directory.
                            continue
                        if file_path.startswith('.') or file_path.find('/.') != -1:
                            # continue when 'file_path' is hidden directory or hidden file.
                            continue

                        if file_name_start_index == -1:
                            yield file_path, file, ''
                        else:
                            yield file_path, file, parent_directory
        except zipfile.BadZipFile:
            raise InvalidPackageException("Bad zip file.")
        except zipfile.LargeZipFile:
            raise InvalidPackageException("Large zip file.")
        except Exception as e:
            raise InvalidPackageException(f'extract_files_gen error -> exception: {e}')



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from enum import Enum, auto
from typing import TYPE_CHECKING, Any, List, Tuple, Optional

from ..base.exception import ExceptionCode, IconServiceBaseException, InvalidRequestException
from ..icon_constant import MAX_EXTERNAL_CALL_COUNT, Revision
from ..utils import to_camel_case, is_lowercase_hex_string, byte_length_of_int

if TYPE_CHECKING:
    from .icon_score_context import IconScoreContextType
    from ..inv.container import Container as INVContainer


def get_input_data_size(revision: int, input_data: Any) -> int:
    """
    Returns size of input data of a transaction

    :param revision: current revision
    :param input_data: input data of transaction
    :return: size of input data
    """
    if revision < Revision.THREE.value:
        return get_data_size_recursively(input_data)

    if revision >= Revision.FOUR.value and input_data is None:
        return 0

    return get_data_size_using_json_dumps(input_data)


def get_data_size_using_json_dumps(data) -> int:
    data = json.dumps(data, ensure_ascii=False, separators=(',', ':'))
    return len(data.encode())


def get_deploy_content_size(revision: int, content: str) -> int:
    """
    Returns size of deploying content.

    If the content is not binary hex string, it raises an exception.

    :param revision: current revision
    :param content: deploying content of transaction
    :return: size of input content
    """
    if revision < Revision.THREE.value:
        return get_data_size_recursively(content)

    if isinstance(content, str) \
            and content.startswith('0x') \
            and is_lowercase_hex_string(content[2:]) \
            and len(content) % 2 == 0:

        return len(content[2:]) // 2
    else:
        raise InvalidRequestException('Invalid content data')


def get_data_size_recursively(data) -> int:
    """
    Deprecated (Only for revision 2 and below).
    Returns size of data(input data or deploying content) by recursive traversal

    :param data: input data or deploying content
    :return: size of data
    """
    size = 0
    if data:
        if isinstance(data, dict):
            for v in data.values():
                size += get_data_size_recursively(v)
        elif isinstance(data, list):
            for v in data:
                size += get_data_size_recursively(v)
        elif isinstance(data, str):
            # If the value is hexstring, it is calculated as bytes otherwise
            # string
            data_body = data[2:] if data.startswith('0x') else data
            if is_lowercase_hex_string(data_body):
                data_body_length = len(data_body)
                size = data_body_length // 2
                if data_body_length % 2 == 1:
                    size += 1
            else:
                size = len(data.encode('utf-8'))
        else:
            # int and bool
            if isinstance(data, int):
                size = byte_length_of_int(data)
    return size


class AutoValueEnum(Enum):
    # noinspection PyMethodParameters
    def _generate_next_value_(name, start, count, last_values):
        # Generates value from the camel-cased name
        return to_camel_case(name.lower())


class StepType(AutoValueEnum):
    DEFAULT = auto()
    CONTRACT_CALL = auto()
    CONTRACT_CREATE = auto()
    CONTRACT_UPDATE = auto()
    CONTRACT_DESTRUCT = auto()
    CONTRACT_SET = auto()
    GET = auto()
    SET = auto()
    REPLACE = auto()
    DELETE = auto()
    INPUT = auto()
    EVENT_LOG = auto()
    API_CALL = auto()


class OutOfStepException(IconServiceBaseException):
    """ An Exception which is thrown when steps are exhausted.
    """

    def __init__(self,
                 step_limit: int,
                 step_used: int,
                 requested_step: int,
                 step_type: StepType):
        """Constructor

        :param step_limit: step limit of the transaction
        :param step_used: used steps in the transaction
        :param requested_step: consuming steps before the exception is thrown
        :param step_type: step type that
        the exception has been thrown when processing
        """
        self.__step_limit: int = step_limit
        self.__step_used = step_used
        self.__requested_step = requested_step
        self.__step_type = step_type

    @property
    def code(self) -> int:
        return ExceptionCode.OUT_OF_STEP

    @property
    def message(self) -> str:
        """
        Returns the exception message
        :return: the exception message
        """
        return f'Out of step: {self.__step_type.value}'

    @property
    def step_limit(self) -> int:
        """
        Returns step limit of the transaction
        :return: step limit of the transaction
        """
        return self.__step_limit

    @property
    def step_used(self) -> int:
        """
        Returns used steps before the exception is thrown in the transaction
        :return: used steps in the transaction
        """
        return self.__step_used

    # TODO unused code
    @property
    def requested_step(self) -> int:
        """
        Returns consuming steps before the exception is thrown.
        :return: Consuming steps before the exception is thrown.
        """
        return self.__requested_step


class StepTracer(object):
    def __init__(self):
        super().__init__()
        self._cumulative_step: int = 0
        self._steps: List[Tuple[StepType, int, int]] = []

    def __str__(self) -> str:
        if len(self._steps) == 0:
            return ""

        lines: List[str] = []

        for i in range(len(self._steps)):
            item: Tuple[StepType, int, int] = self._steps[i]
            # index | stepType | step | cumulativeStep
            lines.append(f"{i:2} | {item[0].name} | {item[1]} | {item[2]}")

        return "\n".join(lines)

    def __len__(self) -> int:
        return len(self._steps)

    @property
    def cumulative_step(self) -> int:
        return self._cumulative_step

    @property
    def steps(self) -> List[Tuple[StepType, int, int]]:
        return self._steps

    def reset(self):
        self._cumulative_step = 0
        self._steps.clear()

    def add(self, step_type: 'StepType', step: int, cumulative_step: int):
        assert cumulative_step == self._cumulative_step + step

        self._cumulative_step += step
        self._steps.append((step_type, step, cumulative_step))


class IconScoreStepCounter(object):
    """ Counts steps in a transaction
    """

    def __init__(self,
                 step_price: int,
                 step_costs: dict,
                 step_limit: int,
                 step_trace_flag: bool = False) -> None:
        """Constructor

        :param step_price: step price
        :param step_costs: a dict of base step costs
        :param step_limit: step limit for current context type
        """

        self._step_price = step_price
        self._step_costs: dict = step_costs
        self._step_limit: int = step_limit
        self._step_used: int = 0
        self._external_call_count: int = 0
        self._max_step_used: int = 0
        self._step_tracer: Optional[StepTracer] = StepTracer() if step_trace_flag else None

    @property
    def step_price(self) -> int:
        """
        Returns the step price
        :return: step price
        """
        return self._step_price

    @property
    def step_limit(self) -> int:
        """
        Returns step limit of the transaction
        :return: step limit of the transaction
        """
        return self._step_limit

    @property
    def step_used(self) -> int:
        """
        Returns used steps in the transaction
        :return: used steps in the transaction
        """
        return max(self._step_used,
                   self._step_costs.get(StepType.DEFAULT, 0))

    @property
    def max_step_used(self) -> int:
        """
        Returns max used steps in the transaction
        :return: max used steps in the transaction
        """
        return self._max_step_used

    @property
    def step_tracer(self) -> Optional[StepTracer]:
        return self._step_tracer

    def apply_step(self, step_type: StepType, count: int) -> int:
        """ Increases steps for given step cost
        """

        if step_type == StepType.CONTRACT_CALL:
            self._external_call_count += 1
            if self._external_call_count > MAX_EXTERNAL_CALL_COUNT:
                raise InvalidRequestException('Too many external calls')

        step: int = self._step_costs.get(step_type, 0) * count

        return self.consume_step(step_type, step)

    def consume_step(self, step_type: StepType, step: int) -> int:
        step_used: int = self._step_used + step

        self._max_step_used = max(self._max_step_used, step_used)

        if step_used > self._step_limit:
            step_used = self._step_used
            self._step_used = self._step_limit
            raise OutOfStepException(
                self._step_limit, step_used, step, step_type)

        self._step_used = step_used

        # Save the step info to StepTracer to trace step cost
        self._trace_step(step_type, step)

        return step_used

    def _trace_step(self, step_type: StepType, step: int):
        if self._step_tracer is not None:
            self._step_tracer.add(step_type, step, self._step_used)

    def get_step_cost(self, step_type: StepType) -> int:
        return self._step_costs.get(step_type, 0)
